import asyncio
from typing import Any
import aiohttp
import feedparser
from housaku.db import db_connection
from housaku.utils import clean_html, console


async def fetch_feed(client: aiohttp.ClientSession, feed_url: str) -> list[Any]:
    resp = await client.get(feed_url)
    if resp.status >= 400:
        raise Exception(f"something went wrong while fetching '{feed_url}'")

    content = await resp.text()
    d = feedparser.parse(content)
    if d.bozo:
        raise Exception(f"failed to parse '{feed_url}'")

    return [entry for entry in d.entries]


async def fetch_post(client: aiohttp.ClientSession, post_url: str) -> str:
    resp = await client.get(post_url)
    if resp.status >= 400:
        raise Exception(f"something went wrong while fetching '{post_url}'")

    html = await resp.text()
    cleaned_html = clean_html(html)
    return cleaned_html


async def index_feeds(sqlite_url: str, feeds: list[str]) -> None:
    async def process_feed(client: aiohttp.ClientSession, feed_url: str):
        try:
            entries = await fetch_feed(client, feed_url)
            for entry in entries:
                entry_link = entry.link
                uri = f"{entry_link}"

                content = await fetch_post(client, entry_link)
                title = entry.get("title", entry_link)

                with db_connection(sqlite_url) as conn:
                    cursor = conn.cursor()
                    cursor.execute(
                        """
                    INSERT INTO documents (uri, title, type, content)
                    VALUES (?, ?, ?, ?)
                    """,
                        (uri, title, "web", content),
                    )
                    console.print(f"[green][Ok][/] indexed '{uri}'.")
        except Exception as e:
            console.print(f"[red][Err][/] {e}")

    async with aiohttp.ClientSession() as client:
        tasks = [process_feed(client, feed) for feed in feeds]
        await asyncio.gather(*tasks)



import asyncio
import concurrent.futures
import os
import subprocess
import textwrap
import urllib.parse
from functools import partial
from multiprocessing import cpu_count
from pathlib import Path
from time import perf_counter
import rich_click as click
from uvicorn import run
from housaku.web import app as web_app
from housaku.tui import app as tui_app
from housaku.db import init_db, db_connection
from housaku.feeds import index_feeds
from housaku.files import (
    list_files,
    index_file,
    COMPLEX_DOCUMENT_FILETYPES,
    PLAIN_TEXT_FILETYPES,
    FILETYPES_SET,
)
from housaku.settings import Settings, config_file_path
from housaku.search import search
from housaku.utils import console

settings = Settings()
init_db(settings.sqlite_url)


@click.group(
    help=settings.description,
    epilog="Check out https://github.com/dnlzrgz/housaku for more details",
    context_settings=dict(
        help_option_names=["-h", "--help"],
    ),
)
@click.version_option(version=settings.version)
def cli() -> None:
    pass


@cli.command(
    name="index",
    short_help="Start indexing documents and posts.",
)
@click.option(
    "-i",
    "--include",
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        readable=True,
    ),
    multiple=True,
    help="Directory to include for indexing.",
)
@click.option(
    "-e",
    "--exclude",
    multiple=True,
    help="Directory or pattern to exclude from indexing.",
)
@click.option(
    "-t",
    "--max-threads",
    type=click.IntRange(min=1),
    default=cpu_count() // 2,
    help="Maximum number of threads to use for indexing (default: half of CPU cores).",
)
def index(include, exclude, max_threads) -> None:
    """
    Index documents and posts from the specified sources in the
    config.toml file.
    """

    merged_include = list(
        set(Path(d) for d in settings.files.include) | {Path(d) for d in include}
    )
    merged_exclude = list(set(settings.files.exclude) | set(exclude))
    urls = settings.feeds.urls

    partial_function = partial(index_file, settings.sqlite_url)

    with console.status(
        "[green]Start indexing... Please, wait a moment.",
        spinner="arrow",
    ) as status:
        try:
            with db_connection(settings.sqlite_url) as conn:
                conn.execute("DELETE FROM documents")
                console.print("[green][Ok][/] Cleaned database.")
        except Exception as e:
            console.print(f"[red][Err][/] Failed to clear database: {e}")
            return

        for dir in merged_include:
            status.update(
                f"[green]Indexing documents from '{dir.name}'... Please wait, this may take a moment.[/]"
            )
            files = list_files(dir, merged_exclude)

            with concurrent.futures.ThreadPoolExecutor(
                max_workers=max_threads
            ) as executor:
                executor.map(partial_function, files)

        status.update(
            "[green]Indexing feeds and posts... Please wait, this may take a moment."
        )
        asyncio.run(index_feeds(settings.sqlite_url, urls))


@cli.command(
    name="web",
    short_help="Launchs Web UI.",
)
@click.option(
    "-p",
    "--port",
    default=4242,
    help="Port.",
)
def start_web(port: int) -> None:
    run(web_app, host="127.0.0.1", port=port)


@cli.command(
    name="tui",
    short_help="Starts Textual TUI.",
)
def start_tui() -> None:
    tui_app(settings).run()


@cli.command(
    name="search",
    short_help="Search for documents and posts.",
)
@click.option(
    "-q",
    "--query",
    prompt="Search terms",
    help="Search terms to find relevant documents.",
)
@click.option(
    "-l",
    "--limit",
    type=int,
    default=20,
    help="Limit the number of documents returned.",
)
def search_documents(query: str, limit: int) -> None:
    """
    Search for documents and posts based on the provided query.
    """

    start_time = perf_counter()

    try:
        results = search(settings.sqlite_url, query, limit)
    except Exception as e:
        console.print(f"[red][Err][/] Something went wrong with your query: {e}")
        return

    if not results:
        console.print("[yellow]No results found.[/]")
        return

    end_time = perf_counter()
    elapsed_time = end_time - start_time

    for uri, title, doc_type, content in results:
        encoded_uri = urllib.parse.quote(uri, safe=":/")
        doc_title = title if title else uri

        truncated_content = textwrap.shorten(content, width=280, placeholder="...")
        emoji = ":globe_with_meridians:"
        link = f"[link={uri}]{doc_title}[/]"
        if doc_type in FILETYPES_SET:
            link = f"[link=file://{encoded_uri}]{doc_title}[/]"

        if doc_type in PLAIN_TEXT_FILETYPES:
            emoji = ":page_facing_up:"
        elif doc_type == "application/pdf":
            emoji = ":scroll:"
        elif doc_type == "application/epub+zip":
            emoji = ":green_book:"
        elif doc_type in COMPLEX_DOCUMENT_FILETYPES:
            emoji = ":bookmark_tabs:"

        console.print(
            f"{emoji} [bold underline]{link}[/]\n{truncated_content}",
            emoji=True,
            justify="left",
            highlight=False,
            end="\n\n",
        )

    console.print(
        f"Search completed in {elapsed_time:.3f}s",
        highlight=False,
    )


@cli.command(
    name="config",
    help="Open the configuration file.",
)
def config() -> None:
    """
    Open the configuration file in the default text editor.
    """

    editor = os.environ.get("EDITOR", None)
    try:
        if editor:
            subprocess.run([editor, str(config_file_path)], check=True)
        else:
            subprocess.run(["open", str(config_file_path)], check=True)
    except Exception as e:
        console.print(f"[red][Err][/] Failed to open the configuration file: {e}")


@cli.command(
    name="vacuum",
    help="Reclaim unused spaced in the database.",
)
def vacuum() -> None:
    """
    Reclaims unused space in the database.
    """

    try:
        with db_connection(settings.sqlite_url) as conn:
            conn.execute("VACUUM")
            console.print("[green][Ok][/] unused space has been reclaimed!")
    except Exception as e:
        console.print(f"[red][Err][/] something went wrong while reclaiming space: {e}")



import sqlite3
from contextlib import contextmanager


def init_db(sqlite_url: str) -> None:
    conn = sqlite3.connect(sqlite_url)
    cursor = conn.cursor()

    cursor.execute("""
    CREATE VIRTUAL TABLE IF NOT EXISTS documents USING fts5 (
        uri, title, type, content, tokenize="porter unicode61"
    );
    """)

    cursor.execute("PRAGMA journal_mode = WAL;")
    cursor.execute("PRAGMA foreign_keys = ON;")
    cursor.execute("PRAGMA synchronous = NORMAL;")
    cursor.execute("PRAGMA busy_timeout = 5000;")
    cursor.execute("PRAGMA temp_store = MEMORY;")
    cursor.execute("PRAGMA cache_size = 2000;")

    cursor.execute("PRAGMA transaction_mode = IMMEDIATE;")

    conn.commit()
    conn.close()


@contextmanager
def db_connection(sqlite_url: str):
    conn = sqlite3.connect(sqlite_url)

    try:
        yield conn
        conn.commit()
    except:
        conn.rollback()
        raise
    finally:
        conn.close()



from rich.console import Console
from selectolax.parser import HTMLParser

console = Console()


def clean_html(html: str) -> str:
    tree = HTMLParser(html)
    for tag in tree.css("script, style, video, img, canvas"):
        tag.decompose()

    text = "".join(node.text(deep=True) for node in tree.css("main"))
    lines = (line.strip() for line in text.splitlines())
    chunks = (phrase.strip() for line in lines for phrase in line.split(" "))
    cleaned_text = " ".join(chunk for chunk in chunks if chunk)

    return cleaned_text
